from .beam import *
from .accelerator import *
from .solver import *

__version__ = '0.0.13'
__doc__ = '''Kapchinscky ENVelope (KENV) -
 solver of the Kapchinsky-Vladimirsky envelope equation'''

__all__ = beam.__all__ + accelerator.__all__ + solver.__all__



# accelerator.py

'''Create an accelerator.
'''

import numpy as np
from scipy import interpolate

__all__ = ['Element',
           'Accelerator',
           'read_elements']


class Element:
    '''Sets one accelerator element.

    Sets one accelerator element with parameters:
    z0 [m] --- element position,
    max_field --- maximum field,
    file_name --- field profile,
    name --- unique element name.
    '''
    def __init__(self,
                 z0: float,
                 max_field: float,
                 file_name: str,
                 name: str):
        self.z0 = z0
        self.max_field = max_field
        self.file_name = file_name
        self.name = name


field_files = {} # buffer for field files

def read_elements(beamline: dict,
                  z:np.arange) -> interpolate.interp1d:
    '''Sews elements into a function of z.

    Sews elements into a function of z with parameters:
    beamline --- set of accelerator elements,
    z [m] --- coordinate.
    '''
    global field_files
    F = 0
    F_dev = 0
    if not beamline:
        z_data = [i/1000 for i in range(1000)]
        F_data = [0 for i in range(1000)]
        f = interpolate.interp1d(
            z_data, F_data,
            fill_value=(0, 0), bounds_error=False
        )
        F = F + f(z)
        F_dev = F
    else:
        for element in beamline.values():
            if not (element.file_name in field_files):
                field_files[element.file_name] = np.loadtxt(element.file_name)
            M = field_files[element.file_name]
            z_data = M[:,0]
            F_data = M[:,1]
            dz = 2*(z_data[-1] - z_data[0])/len(z_data)

            f = interpolate.interp1d(
                element.z0+z_data, element.max_field*F_data, kind='cubic',
                fill_value=(0, 0), bounds_error=False
            )
            F = F + f(z)

            z_data_dev = z_data
            F_data_dev = np.gradient(F_data, dz)
            f_dev = interpolate.interp1d(
                element.z0+z_data_dev, -element.max_field*F_data_dev, kind='cubic',
                fill_value=(0, 0), bounds_error=False
            )
            F_dev = F_dev + f_dev(z)
    F = interpolate.interp1d(z, F, kind='cubic', fill_value=(0, 0), bounds_error=False)
    F_dev = interpolate.interp1d(z, F_dev, kind='cubic', fill_value=(0, 0), bounds_error=False)
    return F, F_dev


class Accelerator:
    '''Create an accelerator.

    Create an accelerator with parameters:
    start [m] --- the beginning of the accelerator,
    stop [m] --- end of the accelerator,
    step [m] --- step method along the accelerator.

    Can add a solenoids, quadrupoles and accelerating modules.

    Accelerator's parameters after compile:
    beamline:
    Bz_beamline, Ez_beamline, Gz_beamline
    function:
    Ez, Bz, Gz
    and
    dEzdz, dBzdz, dGzdz

    '''
    def __init__(self,
                 start: float,
                 stop: float,
                 step: float):
        self.start = start
        self.stop = stop
        self.step = step
        self.z = self.parameter = np.arange(start, stop, step)
        self.Bz_beamline = {}
        self.Ez_beamline = {}
        self.Gz_beamline = {}
        self.Bz = interpolate.interp1d
        self.Ez = interpolate.interp1d
        self.Gz = interpolate.interp1d
        self.dBzdz = interpolate.interp1d
        self.dEzdz = interpolate.interp1d
        self.dGzdz = interpolate.interp1d

    def add_solenoid(self,
                     name: str,
                     center: float,
                     max_field: float,
                     file_name: str ) -> None:
        '''
        Creates a solenoid in the accelerator.

        Creates a solenoid in the accelerator with parameters:
        name --- solenoid's id,
        centert [m] --- solenoid's center,
        max field [T] --- solenoid's maximum field,
        file name --- experimental profile of the Bz field,
        '''
        self.Bz_beamline[name] = Element(center, max_field, file_name, name)

    def add_accel(self,
                     name: str,
                     center: float,
                     max_field: float,
                     file_name: str ) -> None:
        '''
        Creates an accelerating module in the accelerator.

        Creates an accelerating module in the accelerator with parameters:
        name --- accelerating module's id,
        centert [m] --- accelerating module's center,
        max field [MV/m] --- accelerating module's maximum field,
        file name --- experimental profile of the Ez field,
        '''
        self.Ez_beamline[name] = Element(center, max_field, file_name, name)

    def add_quadrupole(self,
                     name: str,
                     center: float,
                     max_field: float,
                     file_name: str ) -> None:
        '''
        Creates a quadrupole in the accelerator.

        Creates a quadrupole in the accelerator with parameters:
        name --- quadrupole's id,
        centert [m] --- quadrupole's center,
        max field [T/m] --- quadrupole's maximum field,
        file name --- experimental profile of the Gz field,
        '''
        self.Gz_beamline[name] = Element(center, max_field, file_name, name)

    def delete_solenoid(self,
                     name: str='all') -> None:
        '''
        Delete a solenoid in the accelerator.

        Delete a solenoid in the accelerator with parameters:
        name --- solenoid's id

        *if there is no name that will be removed all
        '''
        if name == 'all':
            self.Bz_beamline = {}
        else:
            self.Bz_beamline.pop(name)

    def delete_accel(self,
                     name: str='all') -> None:
        '''
        Delete a accelerating module in the accelerator.

        Delete a accelerating module in the accelerator with parameters:
        name --- quadrupole's id

        *if there is no name that will be removed all
        '''
        if name == 'all':
            self.Ez_beamline = {}
        else:
            self.Ez_beamline.pop(name)

    def delete_quadrupole(self,
                     name: str='all') -> None:
        '''
        Delete a quadrupole in the accelerator.

        Delete a quadrupole in the accelerator with parameters:
        name --- quadrupole's id

        *if there is no name that will be removed all
        '''
        if name == 'all':
            self.Gz_beamline = {}
        else:
            self.Gz_beamline.pop(name)


    def compile(self) -> None:
        '''Compilation of the accelerator.
        '''
        self.Bz, self.dBzdz = read_elements(self.Bz_beamline, self.parameter)
        self.Ez, self.dEzdz = read_elements(self.Ez_beamline, self.parameter)
        self.Gz, self.dGzdz = read_elements(self.Gz_beamline, self.parameter)



    def __str__(self):
        string = 'Accelerator structure.\n'
        string += '\tSolenoids:\n'
        for element in self.Bz_beamline.values():
            string +="\t[ %.5f m, %.5f T, '%s', '%s'],\n" % (element.z0, element.max_field, element.file_name, element.name)
        string += '\tAccelerating modules:\n'
        for element in self.Ez_beamline.values():
            string +="\t[ %.5f m, %.5f Mv/m, '%s', '%s'],\n" % (element.z0, element.max_field, element.file_name, element.name)
        string += '\tQuadrupoles:\n'
        for element in self.Gz_beamline.values():
            string +="\t[ %.5f m, %.5f T/m, '%s', '%s'],\n" % (element.z0, element.max_field, element.file_name, element.name)
        return string

    add_sol = add_new_solenoid = add_solenoid
    add_acc = add_new_accel = add_accel
    add_quad = add_new_quadrupole = add_quadrupole

    del_sol = del_solenoid = delete_solenoid
    del_acc = del_accel = delete_accel
    del_quad = del_quadrupole = delete_quadrupole



# solver.py
'''Simulation of the envelope beam in the accelerator.
'''
import numpy as np

from .beam import *
from .accelerator import *
from scipy import interpolate, integrate
from scipy.integrate import solve_ivp

__all__ = ['Sim',
           'Simulation']

class Simulation:
    '''Simulation of the envelope beam in the accelerator.

    basic parameters after track:
    envelope_x and anvelope_y,
    centroid_x and centroid_y,
    larmor_angle
    '''
    mass_rest_electron = 0.511
    clight = 299792458
    alfven_current = 17000

    def __init__(self,
             beam,
             accelerator):

        self.current = beam.current
        self.energy = beam.energy
        self.x = beam.x
        self.y = beam.y
        self.radius_x = beam.radius_x
        self.radius_y = beam.radius_y
        self.xp = beam.xp
        self.yp = beam.yp
        self.normalized_emittance_x = beam.normalized_emittance_x
        self.normalized_emittance_y = beam.normalized_emittance_y
        self.phase = beam.larmor_angle

        self.start = accelerator.start
        self.stop = accelerator.stop
        self.step = accelerator.step

        self.z = self.parameter = accelerator.parameter
        self.Bz = accelerator.Bz
        self.Ez = accelerator.Ez
        self.Gz = accelerator.Gz
        self.dBzdz = accelerator.dBzdz
        self.dEzdz = accelerator.dEzdz
        self.dGzdz = accelerator.dGzdz
        self.gamma_0 = self.energy/self.mass_rest_electron + 1
        self.gamma = self.gamma_0 + integrate.cumtrapz(-self.Ez(self.parameter)/self.mass_rest_electron, self.parameter)
        self.gamma = interpolate.interp1d(self.parameter[1:], self.gamma, fill_value=(self.gamma[0], self.gamma[-1]), bounds_error=False)

        self.envelope_x = []
        self.envelope_y = []

        self.centroid_x = []
        self.centroid_y = []
        self.larmor_angle = []

    def track(self, rtol:float = 1e-6):
        '''Tracking!
        '''

        def dXdz_beam(z:np.arange, X:list,
                 dz:float=self.step,
                 gamma:interpolate.interp1d=self.gamma,
                 Bz:interpolate.interp1d=self.Bz,
                 Ez:interpolate.interp1d=self.Ez,
                 Gz:interpolate.interp1d=self.Gz,
                 dBzdz:interpolate.interp1d=self.dBzdz,
                 dEzdz:interpolate.interp1d=self.dEzdz,
                 dGzdz:interpolate.interp1d=self.dGzdz,
                 beam_current:float=self.current,
                 emitt_n_x:float=self.normalized_emittance_x,
                 emitt_n_y:float=self.normalized_emittance_y,
                 c:float=self.clight,
                 mc:float=self.mass_rest_electron,
                 alfven_current:float=self.alfven_current) -> list:
            '''Located derivative for further integration.
            '''
            sigma_x = X[0]
            x = X[1]
            sigma_y = X[2]
            y = X[3]

            g = gamma(z)
            dgdz = Ez(z) / mc

            d2gdz2 = dEzdz(z) / mc

            beta = np.sqrt(1 - 1 / (g * g))

            K_s = ( c * Bz(z) / (2*mc*1e6) / (g * beta))**2
            K_q = ( c * Gz(z) / (mc*1e6)) / (g * beta)
            K_x = K_s + K_q
            K_y = K_s - K_q

            emitt_x = emitt_n_x / (g * beta)
            emitt_y = emitt_n_y / (g * beta)

            P = 2 * beam_current / (alfven_current * beta * beta * beta * g * g * g)

            dsigma_xdz = x
            dxdz = 2 * P / (sigma_x + sigma_y) + emitt_x * emitt_x / (sigma_x * sigma_x * sigma_x) - K_x * sigma_x - \
                   dgdz * x / (beta * beta * g) - d2gdz2 * sigma_x/(2 * beta * beta * g)
            dsigma_ydz = y
            dydz = 2 * P/(sigma_x + sigma_y) + emitt_y * emitt_y / (sigma_y * sigma_y * sigma_y) - K_y * sigma_y - \
                   dgdz * y/(beta * beta * g) - d2gdz2 * sigma_y/(2 * beta * beta * g)


            return [dsigma_xdz, dxdz, dsigma_ydz, dydz]

        def dXdz_centroid(z:np.arange, X:list,
                 dz:float=self.step,
                 gamma:interpolate.interp1d=self.gamma,
                 Bz:interpolate.interp1d=self.Bz,
                 Ez:interpolate.interp1d=self.Ez,
                 Gz:interpolate.interp1d=self.Gz,
                 dBzdz:interpolate.interp1d=self.dBzdz,
                 dEzdz:interpolate.interp1d=self.dEzdz,
                 dGzdz:interpolate.interp1d=self.dGzdz,
                 beam_current:float=.0e0,
                 emitt_n_x:float=.0e0,
                 emitt_n_y:float=.0e0,
                 c:float=self.clight,
                 mc:float=self.mass_rest_electron,
                 alfven_current:float=self.alfven_current,
                 rtol:float=rtol) -> list:
            '''Located derivative for further integration.
            '''
            sigma_x = X[0]
            x = X[1]
            sigma_y = X[2]
            y = X[3]
            phi = X[4]

            g = gamma(z)
            dgdz = Ez(z) / mc
            d2gdz2 = dEzdz(z) / mc

            beta = np.sqrt(1 - 1 / (g * g))

            K_s = ( c * Bz(z) / (2*mc*1e6) / (g * beta))**2
            K_q = ( c * Gz(z) / (mc*1e6)) / (g * beta)
            K_x = K_s + K_q
            K_y = K_s - K_q

            dsigma_xdz = x
            dxdz =  - K_x * sigma_x - dgdz * x / (beta * beta * g) - d2gdz2 * sigma_x/(2 * beta * beta * g)
            dsigma_ydz = y
            dydz = - K_y * sigma_y - dgdz * y/(beta * beta * g) - d2gdz2 * sigma_y/(2 * beta * beta * g)
            dphidz = -K_s**0.5

            return [dsigma_xdz, dxdz, dsigma_ydz, dydz, dphidz]


        X0_beam = np.array([self.radius_x, self.xp, self.radius_y, self.yp])
        X0_centroid = np.array([self.x, 0, self.y, 0, self.phase])

        beam = solve_ivp(dXdz_beam, t_span=[self.parameter[0],self.parameter[-1]], y0=X0_beam, t_eval=self.parameter, rtol=rtol).y
        centroid = solve_ivp(dXdz_centroid, t_span=[self.parameter[0],self.parameter[-1]], y0=X0_centroid, t_eval=self.parameter, rtol=rtol).y

        self.envelope_x = beam[0,:]
        self.envelope_y = beam[2,:]

        self.centroid_x = centroid[0,:]*np.cos(centroid[4,:]) + centroid[2,:]*np.sin(centroid[4,:])
        self.centroid_y = -centroid[0,:]*np.sin(centroid[4,:]) + centroid[2,:]*np.cos(centroid[4,:])
        self.larmor_angle = centroid[4,:]

Sim = Simulation



# beam.py

'''Creating an electron beam.
'''

__all__ = ['Beam']

class Beam:
    '''Creating an electron beam.

    Creating a round electron beam with parameters:
    current [A],
    energy [MeV],
    radius [m],
    rp [m],
    normalized emittance [m*rad],

    Creating an elliptical electron beam with parameters:
    radius x [m],
    radius y [m],
    xp [rad],
    yp [rad],
    normalized emittance x [m*rad],
    normalized emittance y [m*rad]

    and with shifted centroid:
    x [m], y [m], larmor_angle [rad]
    '''

    def __init__(self,
                 current: float=.0e0,
                 energy: float=.0e0,
                 radius: float=.0e0,
                 radius_x: float=.0e0,
                 radius_y: float=.0e0,
                 rp: float=.0e0,
                 xp: float=.0e0,
                 yp: float=.0e0,
                 normalized_emittance: float=.0e0,
                 normalized_emittance_x: float=.0e0,
                 normalized_emittance_y: float=.0e0,
                 x: float=.0e0,
                 y: float=.0e0,
                 larmor_angle: float=.0e0):

        self.current = current
        self.energy = energy
        self.radius = radius
        self.radius_x = radius_x
        self.radius_y = radius_y
        self.rp = rp
        self.xp = xp
        self.yp = yp
        self.normalized_emittance = normalized_emittance
        self.normalized_emittance_x = normalized_emittance_x
        self.normalized_emittance_y = normalized_emittance_y
        if radius != .0e0:
            self.radius_x = radius
            self.radius_y = radius
        if rp !=.0e0:
            self.xp = rp
            self.yp = rp
        if normalized_emittance !=.0e0:
            self.normalized_emittance_x = normalized_emittance
            self.normalized_emittance_y = normalized_emittance
        self.particles = {}
        self.x = x
        self.y = y
        self.larmor_angle = larmor_angle

    def __str__(self):
            return 'Beam parameters:' + '\n' \
                    +'\tCurrent\t%0.1f A'%(self.current) + '\n' \
                    +'\tEnergy\t%0.1f MeV'%(self.energy) + '\n' \
                    +'\tx position\t%0.1f mm'%(self.x*1e3) + '\n' \
                    +'\ty position\t%0.1f mm'%(self.y*1e3) + '\n' \
                    +'\tRadius y\t%0.1f mm'%(self.radius_y*1e3) + '\n' \
                    +'\tRadius x\t%0.1f mm'%(self.radius_x*1e3) + '\n' \
                    +'\trp x\t%0.1f mrad'%(self.xp*1e3) + '\n' \
                    +'\trp y\t%0.1f mrad'%(self.yp*1e3) + '\n' \
                    +'\tLarmor larmor_angle\t%0.1f mrad'%(self.larmor_angle*1e3) + '\n' \
                    +'\tNormalized emittans x\t%0.1f mm*mrad'%\
                    (self.normalized_emittance_x*1e6) + '\n' \
                    +'\tNormalized emittans y\t%0.1f mm*mrad'%\
                    (self.normalized_emittance_y*1e6) + '\n' \
